@@ -21,7 +21,14 @@
 
 """separate file for database operations"""
 
-from pysqlite2 import dbapi2 as sqlite
+try:
+    import sqlite3 as sqlite
+except ImportError:
+    try:
+        from pysqlite2 import dbapi2 as sqlite
+    except ImportError:
+        print "Error: Neither sqlite3 nor pysqlite2 found"
+        raise
 import os, time
 import datetime
 import hamster
@@ -37,22 +44,39 @@ class Storage(hamster.storage.Storage):
         return self.fetchall("SELECT * FROM categories ORDER BY category_order")
 
     def __change_category(self, id, category_id):
-        query = "SELECT max(activity_order) + 1 FROM activities WHERE category_id = ?"
-        max_order = self.fetchone(query, (category_id, ))[0] or 1
+        # first check if we don't have an activity with same name before us
+        activity = self.fetchone("select name from activities where id = ?", (id, ))
+        existing_id = self.__get_activity_by_name(activity['name'], category_id)
         
-        statement = """
-                   UPDATE activities 
-                      SET category_id = ?, activity_order = ?
-                    WHERE id = ?
-        """
+        if existing_id: #ooh, we have something here!
+            # first move all facts that belong to movable activity to the new one
+            update = """
+                       UPDATE facts
+                          SET activity_id = ?
+                        WHERE activity_id = ?
+            """
+            
+            self.execute(update, (existing_id, id))
+            
+            # and now get rid of our friend
+            self.__remove_activity(id)
         
-        self.execute(statement, (category_id, max_order, id))
+        else: #just moving        
+            query = "SELECT max(activity_order) + 1 FROM activities WHERE category_id = ?"
+            max_order = self.fetchone(query, (category_id, ))[0] or 1
+            
+            statement = """
+                       UPDATE activities 
+                          SET category_id = ?, activity_order = ?
+                        WHERE id = ?
+            """
+            
+            self.execute(statement, (category_id, max_order, id))
     
-    def __insert_category(self, name):
+    def __add_category(self, name):
         new_rec = self.fetchone("select max(id) +1, max(category_order) + 1  from categories")
-        id, order = 1, 1
-        if new_rec[0] and new_rec[1]: # handle case when we have no categories at all
-            id, order = new_rec[0], new_rec[1]
+
+        id, order = new_rec[0] or 1, new_rec[1] or 1
 
         query = """
                    INSERT INTO categories (id, name, category_order)
@@ -89,16 +113,55 @@ class Storage(hamster.storage.Storage):
             
         
         
-    def __get_activity_by_name(self, name):
+    def __get_activity_by_name(self, name, category_id = None):
         """get most recent, preferably not deleted activity by it's name"""
+        
+        if category_id:
+            query = """
+                       SELECT id, deleted from activities 
+                        WHERE lower(name) = lower(?)
+                          AND category_id = ?
+                     ORDER BY deleted, id desc
+                        LIMIT 1
+            """
+            
+            res = self.fetchone(query, (name, category_id))
+        else:
+            query = """
+                       SELECT id, deleted from activities 
+                        WHERE lower(name) = lower(?)
+                     ORDER BY deleted, id desc
+                        LIMIT 1
+            """
+            
+            res = self.fetchone(query, (name,))
+        
+        if res:
+            # if the activity was marked as deleted, ressurect on first call
+            # and put in the unsorted category
+            if res['deleted']:
+                update = """
+                            UPDATE activities
+                               SET deleted = null, category_id = -1
+                             WHERE id = ?
+                        """
+                self.execute(update, (res['id'], ))
+            
+            return res['id']
+        
+        return None
+
+    def __get_category_by_name(self, name):
+        """returns category by it's name"""        
+
         query = """
-                   SELECT id from activities 
+                   SELECT id from categories
                     WHERE lower(name) = lower(?)
-                 ORDER BY deleted, id desc
+                 ORDER BY id desc
                     LIMIT 1
         """
-        
-        res = self.fetchone(query, (name,))
+            
+        res = self.fetchone(query, (name, ))
         
         if res:
             return res['id']
@@ -109,6 +172,7 @@ class Storage(hamster.storage.Storage):
         query = """SELECT a.id AS id,
                           a.start_time AS start_time,
                           a.end_time AS end_time,
+                          a.description as description,
                           b.name AS name, b.id as activity_id
                      FROM facts a
                 LEFT JOIN activities b ON a.activity_id = b.id
@@ -121,6 +185,7 @@ class Storage(hamster.storage.Storage):
                    SELECT a.id AS id,
                           a.start_time AS start_time,
                           a.end_time AS end_time,
+                          a.description as description,
                           b.name AS name, b.id as activity_id
                      FROM facts a
                 LEFT JOIN activities b ON a.activity_id = b.id
@@ -130,23 +195,44 @@ class Storage(hamster.storage.Storage):
         """
         return self.fetchone(query, (dt.date.today(), ))
 
-    def __touch_fact(self, activity, end_time = None):
-        id = activity['id']
-        query = """
-                   UPDATE facts
-                      SET end_time = ?
-                    WHERE id = ?
-        """
-        self.execute(query, (end_time, id))
+    def __touch_fact(self, fact, end_time):
+        # tasks under one minute do not count
+        if end_time - fact['start_time'] < datetime.timedelta(minutes = 1):
+            self.__remove_fact(fact['id'])
+        else:
+            query = """
+                       UPDATE facts
+                          SET end_time = ?
+                        WHERE id = ?
+            """
+            self.execute(query, (end_time, fact['id']))
 
     def __add_fact(self, activity_name, start_time = None, end_time = None):
         start_time = start_time or datetime.datetime.now()
         
-        # try to lookup activity by it's name in db. active ones have priority
-        activity_id = self.__get_activity_by_name(activity_name)
+        #see if we have description of activity somewhere here (delimited by comma)
+        description = None
+        if activity_name.find(",") > 0:
+            activity_name, description = activity_name.split(",", 1)
+            description = description.strip()
+        
+        # now check if maybe there is also a category
+        category_id = None
+        if activity_name.find("@") > 0:
+            #at symbol marks category
+            activity_name, category_name = activity_name.split("@", 1)
+            
+            if category_name:
+                category_name = category_name.strip()
+                category_id = self.__get_category_by_name(category_name)
+                if not category_id:
+                    category_id = self.__add_category(category_name)
+        
+        # try to find activity
+        activity_id = self.__get_activity_by_name(activity_name, category_id)
 
         if not activity_id:
-            activity_id = self.__insert_activity(activity_name)
+            activity_id = self.__add_activity(activity_name, category_id)
 
 
         # now fetch facts for the specified day and check if we have to
@@ -185,7 +271,12 @@ class Storage(hamster.storage.Storage):
         if day_facts:
             last_fact = day_facts[-1]
             
-            if last_fact['end_time'] == None and start_time > last_fact['start_time']:
+            if last_fact['end_time'] == None and last_fact['start_time'] < start_time:
+                #if this is the same, ongoing activity, then there is no need to create another one
+                if last_fact['activity_id'] == activity_id:
+                    return last_fact
+                
+                
                 delta = (start_time - last_fact['start_time'])
             
                 if 60 >= delta.seconds >= 0:
@@ -217,19 +308,11 @@ class Storage(hamster.storage.Storage):
 
 
         # finally add the new entry
-        if end_time:
-            insert = """
-                        INSERT INTO facts (activity_id, start_time, end_time)
-                                   VALUES (?, ?, ?)
-            """
-            self.execute(insert, (activity_id, start_time, end_time))
-        else:
-            insert = """
-                        INSERT INTO facts (activity_id, start_time)
-                                   VALUES (?, ?)
-            """
-            self.execute(insert, (activity_id, start_time))
-
+        insert = """
+                    INSERT INTO facts (activity_id, start_time, end_time, description)
+                               VALUES (?, ?, ?, ?)
+        """
+        self.execute(insert, (activity_id, start_time, end_time, description))
 
         fact_id = self.fetchone("select max(id) as max_id from facts")['max_id']
         
@@ -241,6 +324,7 @@ class Storage(hamster.storage.Storage):
                    SELECT a.id AS id,
                           a.start_time AS start_time,
                           a.end_time AS end_time,
+                          a.description as description,
                           b.name AS name, b.id as activity_id,
                           coalesce(c.name, ?) as category, c.id as category_id
                      FROM facts a
@@ -260,10 +344,42 @@ class Storage(hamster.storage.Storage):
         """
         self.execute(query, (fact_id,))
 
+    def __get_activities(self, category_id = None):
+        """returns list of activities, if category is specified, order by name
+           otherwise - by activity_order"""
+        if category_id:
+            query = """
+                       SELECT a.*, b.name as category
+                         FROM activities a
+                    LEFT JOIN categories b on coalesce(b.id, -1) = a.category_id
+                        WHERE category_id = ?
+                          AND deleted is null
+            """
+            
+            # unsorted entries we sort by name - others by ID
+            if category_id == -1:
+                query += "ORDER BY lower(a.name)"
+            else:
+                query += "ORDER BY a.activity_order"
+                
+            activities = self.fetchall(query, (category_id, ))
+            
+        else:
+            query = """
+                       SELECT a.*, b.name as category
+                         FROM activities a
+                    LEFT JOIN categories b on coalesce(b.id, -1) = a.category_id
+                        WHERE deleted is null
+                     ORDER BY lower(a.name)
+            """
+            activities = self.fetchall(query)
+            
+        return activities
+
     def __get_sorted_activities(self):
         """returns list of acitivities that have categories"""
         query = """
-                   SELECT a.*, b.category_order
+                   SELECT a.*, b.name as category, b.category_order
                      FROM activities a
                 LEFT JOIN categories b on coalesce(b.id, -1) = a.category_id
                     WHERE a.category_id > -1
@@ -272,35 +388,19 @@ class Storage(hamster.storage.Storage):
         """
         return self.fetchall(query)
         
+    def __get_autocomplete_activities(self):
+        """returns list of activities for autocomplete,
+           activity names converted to lowercase"""
+
+        query = """
+                   SELECT lower(a.name) as name, b.name as category
+                     FROM activities a
+                LEFT JOIN categories b on coalesce(b.id, -1) = a.category_id
+                    WHERE deleted is null
+                 ORDER BY lower(a.name)
+        """
+        activities = self.fetchall(query)
         
-    def __get_activities(self, category_id = None):
-        """returns list of activities, if category is specified, order by name
-           otherwise - by activity_order"""
-        if category_id:
-            query = """
-                       SELECT *
-                         FROM activities
-                        WHERE category_id = ?
-                          AND deleted is null
-            """
-            
-            # unsorted entries we sort by name - others by ID
-            if category_id == -1:
-                query += "ORDER BY lower(name)"
-            else:
-                query += "ORDER BY activity_order"
-                
-            activities = self.fetchall(query, (category_id, ))
-            
-        else:
-            query = """
-                       SELECT *
-                         FROM activities
-                        WHERE deleted is null
-                     ORDER BY lower(name)
-            """
-            activities = self.fetchall(query)
-            
         return activities
 
     def __remove_activity(self, id):
@@ -333,11 +433,16 @@ class Storage(hamster.storage.Storage):
         self.execute("update activities set activity_order = ? where id = ?", (priority1, id2) )
         self.execute("update activities set activity_order = ? where id = ?", (priority2, id1) )
 
-    def __insert_activity(self, name, category_id = -1):
+    def __add_activity(self, name, category_id = None):
+        # first check that we don't have anything like that yet
+        activity_id = self.__get_activity_by_name(name, category_id)
+        if activity_id:
+            return activity_id
+        
+        #now do the create bit
+        category_id = category_id or -1
         new_rec = self.fetchone("select max(id) + 1 , max(activity_order) + 1  from activities")
-        new_id, new_order = 0, 0
-        if new_rec[0] and new_rec[1]: # handle case when we have no activities at all
-            new_id, new_order = new_rec[0], new_rec[1]
+        new_id, new_order = new_rec[0] or 1, new_rec[1] or 1
 
         query = """
                    INSERT INTO activities (id, name, category_id, activity_order)
@@ -413,6 +518,7 @@ class Storage(hamster.storage.Storage):
         
         """upgrade DB to hamster version"""
         version = self.fetchone("SELECT version FROM version")["version"]
+        current_version = 5
 
         if version < 2:
             """moving from fact_date, fact_time to start_time, end_time"""
@@ -573,11 +679,17 @@ class Storage(hamster.storage.Storage):
 
             self.execute("DROP TABLE activities")
             self.execute("ALTER TABLE activities_new RENAME TO activities")
-            
-
-        #lock down current version
-        self.execute("UPDATE version SET version = 4")
         
+        if version < 5:
+            self.execute("ALTER TABLE facts add column description varchar2")
+
+
+        # at the happy end, update version number 
+        if version < current_version:
+            #lock down current version
+            self.execute("UPDATE version SET version = %d" % current_version) 
+        
+
         
         """we start with an empty database and then populate with default
            values. This way defaults can be localized!"""
@@ -585,13 +697,13 @@ class Storage(hamster.storage.Storage):
         category_count = self.fetchone("select count(*) from categories")[0]
         
         if category_count == 0:
-            work_cat_id = self.__insert_category(_(work_category["name"]))
+            work_cat_id = self.__add_category(_(work_category["name"]))
             for entry in work_category["entries"]:
-                self.__insert_activity(_(entry), work_cat_id)
+                self.__add_activity(_(entry), work_cat_id)
         
-            nonwork_cat_id = self.__insert_category(_(nonwork_category["name"]))
+            nonwork_cat_id = self.__add_category(_(nonwork_category["name"]))
             for entry in nonwork_category["entries"]:
-                self.__insert_activity(_(entry), nonwork_cat_id)
+                self.__add_activity(_(entry), nonwork_cat_id)
         
         
         
